@@ -26,7 +26,7 @@
 #include <glib/gstdio.h>
 #include <polkit/polkit.h>
 #include <dbus/dbus-glib.h>
-#include <gsettings-desktop-schemas/gdesktop-enums.h>
+#include <gdesktop-enums.h>
 
 #include <cups/cups.h>
 
@@ -73,6 +73,8 @@ struct _CcPrintersPanelPrivate
 
   GPermission *permission;
 
+  GSettings *lockdown_settings;
+
   PpNewPrinterDialog *pp_new_printer_dialog;
 
   gpointer dummy;
@@ -149,6 +151,12 @@ cc_printers_panel_dispose (GObject *object)
       priv->builder = NULL;
     }
 
+  if (priv->lockdown_settings)
+    {
+      g_object_unref (priv->lockdown_settings);
+      priv->lockdown_settings = NULL;
+    }
+
   G_OBJECT_CLASS (cc_printers_panel_parent_class)->dispose (object);
 }
 
@@ -176,6 +184,16 @@ cc_printers_panel_class_finalize (CcPrintersPanelClass *klass)
 {
 }
 
+enum
+{
+  NOTEBOOK_INFO_PAGE = 0,
+  NOTEBOOK_JOBS_PAGE,
+  NOTEBOOK_OPTIONS_PAGE,
+  NOTEBOOK_NO_PRINTERS_PAGE,
+  NOTEBOOK_NO_CUPS_PAGE,
+  NOTEBOOK_N_PAGES
+};
+
 enum
 {
   PRINTER_ID_COLUMN,
@@ -183,7 +201,6 @@ enum
   PRINTER_PAUSED_COLUMN,
   PRINTER_DEFAULT_ICON_COLUMN,
   PRINTER_ICON_COLUMN,
-  PRINTER_IS_SEPARATOR_COLUMN,
   PRINTER_N_COLUMNS
 };
 
@@ -198,6 +215,7 @@ printer_selection_changed_cb (GtkTreeSelection *selection,
   GtkWidget              *widget;
   gboolean                test_page_command_available = FALSE;
   gboolean                sensitive;
+  gboolean                is_local = TRUE;
   gchar                  *printer_make_and_model = NULL;
   gchar                  *printer_model = NULL;
   gchar                  *reason = NULL;
@@ -206,8 +224,10 @@ printer_selection_changed_cb (GtkTreeSelection *selection,
   gchar                 **printer_reasons = NULL;
   gchar                  *marker_types = NULL;
   gchar                  *printer_name = NULL;
+  gchar                  *printer_type = NULL;
   gchar                  *active_jobs = NULL;
   gchar                  *supply_type = NULL;
+  gchar                  *printer_uri = NULL;
   gchar                  *location = NULL;
   gchar                  *status = NULL;
   guint                   num_jobs;
@@ -292,6 +312,11 @@ printer_selection_changed_cb (GtkTreeSelection *selection,
       priv->current_dest < priv->num_dests &&
       priv->dests != NULL)
     {
+      widget = (GtkWidget*)
+        gtk_builder_get_object (priv->builder, "notebook");
+      if (gtk_notebook_get_current_page (GTK_NOTEBOOK (widget)) >= NOTEBOOK_NO_PRINTERS_PAGE)
+        gtk_notebook_set_current_page (GTK_NOTEBOOK (widget), NOTEBOOK_INFO_PAGE);
+
       for (i = 0; i < priv->dests[id].num_options; i++)
         {
           if (g_strcmp0 (priv->dests[id].options[i].name, "printer-location") == 0)
@@ -306,6 +331,10 @@ printer_selection_changed_cb (GtkTreeSelection *selection,
             printer_commands = priv->dests[priv->current_dest].options[i].value;
           else if (g_strcmp0 (priv->dests[priv->current_dest].options[i].name, "printer-make-and-model") == 0)
             printer_make_and_model = priv->dests[priv->current_dest].options[i].value;
+          else if (g_strcmp0 (priv->dests[priv->current_dest].options[i].name, "printer-uri-supported") == 0)
+            printer_uri = priv->dests[priv->current_dest].options[i].value;
+          else if (g_strcmp0 (priv->dests[priv->current_dest].options[i].name, "printer-type") == 0)
+            printer_type = priv->dests[priv->current_dest].options[i].value;
         }
 
       if (priv->dest_model_names[priv->current_dest] == NULL)
@@ -466,6 +495,44 @@ printer_selection_changed_cb (GtkTreeSelection *selection,
         gtk_label_set_text (GTK_LABEL (widget), EMPTY_TEXT);
 
 
+      widget = (GtkWidget*)
+        gtk_builder_get_object (priv->builder, "printer-ip-address-label");
+
+      if (printer_type)
+        {
+          cups_ptype_t type;
+          type = atoi (printer_type);
+          is_local = !(type & (CUPS_PRINTER_REMOTE | CUPS_PRINTER_IMPLICIT));
+        }
+
+      if (is_local)
+        {
+          gtk_label_set_text (GTK_LABEL (widget), "localhost");
+        }
+      else
+        {
+          if (printer_uri)
+            {
+              char scheme[HTTP_MAX_URI],
+              userpass[HTTP_MAX_URI],
+              server[HTTP_MAX_URI],
+              resource[HTTP_MAX_URI];
+              int port;
+
+              httpSeparateURI(HTTP_URI_CODING_ALL, printer_uri, scheme, sizeof(scheme), userpass,
+                              sizeof(userpass), server, sizeof(server), &port, resource,
+                              sizeof(resource));
+
+              if (server[0] != '\0')
+                gtk_label_set_text (GTK_LABEL (widget), server);
+              else
+                gtk_label_set_text (GTK_LABEL (widget), EMPTY_TEXT);
+            }
+          else
+            gtk_label_set_text (GTK_LABEL (widget), EMPTY_TEXT);
+        }
+
+
       widget = (GtkWidget*)
         gtk_builder_get_object (priv->builder, "printer-disable-switch");
 
@@ -550,9 +617,27 @@ printer_selection_changed_cb (GtkTreeSelection *selection,
       widget = (GtkWidget*)
         gtk_builder_get_object (priv->builder, "printer-jobs-button");
       gtk_widget_set_sensitive (widget, TRUE);
+
+      widget = (GtkWidget*)
+        gtk_builder_get_object (priv->builder, "printer-icon");
+      gtk_widget_set_sensitive (widget, TRUE);
     }
   else
     {
+      http_t *http = NULL;
+
+      widget = (GtkWidget*)
+        gtk_builder_get_object (priv->builder, "notebook");
+
+      http = httpConnectEncrypt (cupsServer (), ippPort (), cupsEncryption ());
+      if (http)
+        {
+          httpClose (http);
+          gtk_notebook_set_current_page (GTK_NOTEBOOK (widget), NOTEBOOK_NO_PRINTERS_PAGE);
+        }
+      else
+        gtk_notebook_set_current_page (GTK_NOTEBOOK (widget), NOTEBOOK_NO_CUPS_PAGE);
+
       widget = (GtkWidget*)
         gtk_builder_get_object (priv->builder, "printer-name-label");
       gtk_label_set_text (GTK_LABEL (widget), "");
@@ -573,6 +658,10 @@ printer_selection_changed_cb (GtkTreeSelection *selection,
         gtk_builder_get_object (priv->builder, "printer-ip-address-label");
       gtk_label_set_text (GTK_LABEL (widget), "");
 
+      widget = (GtkWidget*)
+        gtk_builder_get_object (priv->builder, "printer-jobs-label");
+      gtk_label_set_text (GTK_LABEL (widget), "");
+
       widget = (GtkWidget*)
         gtk_builder_get_object (priv->builder, "print-test-page-button");
       gtk_widget_set_sensitive (widget, FALSE);
@@ -584,6 +673,10 @@ printer_selection_changed_cb (GtkTreeSelection *selection,
       widget = (GtkWidget*)
         gtk_builder_get_object (priv->builder, "printer-jobs-button");
       gtk_widget_set_sensitive (widget, FALSE);
+
+      widget = (GtkWidget*)
+        gtk_builder_get_object (priv->builder, "printer-icon");
+      gtk_widget_set_sensitive (widget, FALSE);
     }
 
   widget = (GtkWidget*)
@@ -599,26 +692,6 @@ printer_selection_changed_cb (GtkTreeSelection *selection,
   gtk_widget_set_sensitive (widget, FALSE);
 }
 
-typedef struct {
-  gint printer_id;
-  gint usage;
-} PrinterUsage;
-
-static gint
-printer_usage_cmp (gconstpointer a,
-                   gconstpointer b)
-{
-  PrinterUsage *x = (PrinterUsage*) a;
-  PrinterUsage *y = (PrinterUsage*) b;
-
-  if (x->usage < y->usage)
-    return 1;
-  else if (x->usage == y->usage)
-    return 0;
-  else
-    return -1;
-}
-
 static void
 actualize_printers_list (CcPrintersPanel *self)
 {
@@ -628,16 +701,12 @@ actualize_printers_list (CcPrintersPanel *self)
   GtkTreeView            *treeview;
   GtkTreeIter             iter;
   cups_job_t             *jobs = NULL;
-  gboolean                has_used_printer = FALSE;
-  gboolean                has_separator = FALSE;
   gboolean                paused = FALSE;
   gboolean                valid = FALSE;
   gchar                  *current_printer_instance = NULL;
   gchar                  *current_printer_name = NULL;
   gchar                  *printer_icon_name = NULL;
   gchar                  *default_icon_name = NULL;
-  GList                  *tmp_list = NULL;
-  GList                  *usages = NULL;
   int                     current_dest = -1;
   int                     i, j;
   int                     num_jobs = 0;
@@ -672,47 +741,28 @@ actualize_printers_list (CcPrintersPanel *self)
                               G_TYPE_STRING,
                               G_TYPE_BOOLEAN,
                               G_TYPE_STRING,
-                              G_TYPE_STRING,
-                              G_TYPE_BOOLEAN);
+                              G_TYPE_STRING);
+
+  if (priv->num_dests == 0)
+    {
+      gtk_list_store_append (store, &iter);
+      gtk_list_store_set (store, &iter,
+                          PRINTER_ID_COLUMN, 0,
+      /* Translators: There are no printers available (none is configured or CUPS is not running) */
+                          PRINTER_NAME_COLUMN, _("No printers available"),
+                          PRINTER_PAUSED_COLUMN, TRUE,
+                          PRINTER_DEFAULT_ICON_COLUMN, NULL,
+                          PRINTER_ICON_COLUMN, NULL,
+                          -1);
+      gtk_widget_set_sensitive (GTK_WIDGET (treeview), FALSE);
+    }
+  else
+    gtk_widget_set_sensitive (GTK_WIDGET (treeview), TRUE);
 
   for (i = 0; i < priv->num_dests; i++)
-    {
-      PrinterUsage *usage = g_new (PrinterUsage, 1);
-      num_jobs = cupsGetJobs (&jobs, priv->dests[i].name, 1, CUPS_WHICHJOBS_ALL);
-      usage->printer_id = i;
-      usage->usage = num_jobs;
-      usages = g_list_append (usages, usage);
-      cupsFreeJobs (num_jobs, jobs);
-    }
-  num_jobs = 0;
-  jobs = NULL;
-
-  usages = g_list_sort (usages, printer_usage_cmp);
-
-  for (tmp_list = usages; tmp_list; tmp_list = tmp_list->next)
     {
       gchar *instance;
 
-      i = ((PrinterUsage*) tmp_list->data)->printer_id;
-      if (((PrinterUsage*) tmp_list->data)->usage == 0)
-        {
-          if (!has_separator && has_used_printer)
-            {
-              has_separator = TRUE;
-              gtk_list_store_append (store, &iter);
-              gtk_list_store_set (store, &iter,
-                                  PRINTER_ID_COLUMN, -1,
-                                  PRINTER_NAME_COLUMN, NULL,
-                                  PRINTER_PAUSED_COLUMN, FALSE,
-                                  PRINTER_DEFAULT_ICON_COLUMN, FALSE,
-                                  PRINTER_ICON_COLUMN, NULL,
-                                  PRINTER_IS_SEPARATOR_COLUMN, TRUE,
-                                  -1);
-            }
-        }
-      else
-        has_used_printer = TRUE;
-
       gtk_list_store_append (store, &iter);
 
       if (priv->dests[i].instance)
@@ -758,14 +808,12 @@ actualize_printers_list (CcPrintersPanel *self)
                           PRINTER_PAUSED_COLUMN, paused,
                           PRINTER_DEFAULT_ICON_COLUMN, default_icon_name,
                           PRINTER_ICON_COLUMN, printer_icon_name,
-                          PRINTER_IS_SEPARATOR_COLUMN, FALSE,
                           -1);
 
       g_free (instance);
       g_free (printer_icon_name);
       g_free (default_icon_name);
     }
-  g_list_free_full (usages, g_free);
 
   gtk_tree_view_set_model (treeview, GTK_TREE_MODEL (store));
 
@@ -849,10 +897,25 @@ set_cell_sensitivity_func (GtkTreeViewColumn *tree_column,
                            GtkTreeIter       *iter,
                            gpointer           func_data)
 {
+  CcPrintersPanelPrivate *priv;
+  CcPrintersPanel        *self = (CcPrintersPanel*) func_data;
   gboolean                paused = FALSE;
 
+  priv = PRINTERS_PANEL_PRIVATE (self);
+
   gtk_tree_model_get (tree_model, iter, PRINTER_PAUSED_COLUMN, &paused, -1);
 
+  if (priv->num_dests == 0)
+    g_object_set (G_OBJECT (cell),
+                  "ellipsize", PANGO_ELLIPSIZE_NONE,
+                  "width-chars", -1,
+                  NULL);
+  else
+    g_object_set (G_OBJECT (cell),
+                  "ellipsize", PANGO_ELLIPSIZE_END,
+                  "width-chars", 18,
+                  NULL);
+
   if (paused)
     g_object_set (cell,
                   "sensitive", FALSE,
@@ -863,23 +926,6 @@ set_cell_sensitivity_func (GtkTreeViewColumn *tree_column,
                   NULL);
 }
 
-static gboolean
-printers_row_separator_func (GtkTreeModel *model,
-                             GtkTreeIter  *iter,
-                             gpointer      data)
-{
-  gboolean separator;
-
-  gtk_tree_model_get (model,
-                      iter,
-                      PRINTER_IS_SEPARATOR_COLUMN,
-                      &separator,
-                      -1);
-
-  return separator;
-}
-
-
 static void
 populate_printers_list (CcPrintersPanel *self)
 {
@@ -925,11 +971,6 @@ populate_printers_list (CcPrintersPanel *self)
                                                      "icon-name", PRINTER_DEFAULT_ICON_COLUMN, NULL);
   gtk_tree_view_column_set_expand (column, FALSE);
   gtk_tree_view_append_column (GTK_TREE_VIEW (treeview), column);
-
-  gtk_tree_view_set_row_separator_func (GTK_TREE_VIEW (treeview),
-                                        printers_row_separator_func,
-                                        NULL,
-                                        NULL);
 }
 
 enum
@@ -1964,6 +2005,129 @@ printer_maintenance_cb (GtkButton *button,
     }
 }
 
+static void
+test_page_cb (GtkButton *button,
+              gpointer   user_data)
+{
+  CcPrintersPanelPrivate  *priv;
+  CcPrintersPanel         *self = (CcPrintersPanel*) user_data;
+  cups_ptype_t             type = 0;
+  const gchar             *printer_type = NULL;
+  gchar                   *printer_name = NULL;
+  gint                     i;
+
+  priv = PRINTERS_PANEL_PRIVATE (self);
+
+  if (priv->current_dest >= 0 &&
+      priv->current_dest < priv->num_dests &&
+      priv->dests != NULL)
+    {
+      printer_name = priv->dests[priv->current_dest].name;
+      printer_type = cupsGetOption ("printer-type",
+                                    priv->dests[priv->current_dest].num_options,
+                                    priv->dests[priv->current_dest].options);
+      if (printer_type)
+        type = atoi (printer_type);
+    }
+
+  if (printer_name)
+    {
+      const gchar  *const dirs[] = { "/usr/share/cups",
+                                     "/usr/local/share/cups",
+                                     NULL };
+      const gchar  *testprint[] = { "%s/data/testprint",
+                                    "%s/data/testprint.ps",
+                                    NULL };
+      const gchar **pattern;
+      const gchar  *datadir = NULL;
+      http_t       *http = NULL;
+      gchar        *printer_uri = NULL;
+      gchar        *filename = NULL;
+      gchar        *resource = NULL;
+      ipp_t        *response;
+      ipp_t        *request;
+
+      if ((datadir = getenv ("CUPS_DATADIR")) != NULL)
+        {
+          for (pattern = testprint; *pattern != NULL; pattern++)
+            {
+              filename = g_strdup_printf (*pattern, datadir);
+              if (g_access (filename, R_OK) == 0)
+                break;
+              else
+                {
+                  g_free (filename);
+                  filename = NULL;
+                }
+            }
+        }
+      else
+        {
+          for (i = 0; (datadir = dirs[i]) != NULL && filename == NULL; i++)
+            {
+              for (pattern = testprint; *pattern != NULL; pattern++)
+                {
+                  filename = g_strdup_printf (*pattern, datadir);
+                  if (g_access (filename, R_OK) == 0)
+                    break;
+                  else
+                    {
+                      g_free (filename);
+                      filename = NULL;
+                    }
+                }
+            }
+        }
+
+      if (filename)
+        {
+          if (type & CUPS_PRINTER_CLASS)
+            {
+              printer_uri = g_strdup_printf ("ipp://localhost/classes/%s", printer_name);
+              resource = g_strdup_printf ("/classes/%s", printer_name);
+            }
+          else
+            {
+              printer_uri = g_strdup_printf ("ipp://localhost/printers/%s", printer_name);
+              resource = g_strdup_printf ("/printers/%s", printer_name);
+            }
+
+          http = httpConnectEncrypt (cupsServer (), ippPort (), cupsEncryption ());
+          if (http)
+            {
+              request = ippNewRequest (IPP_PRINT_JOB);
+              ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_URI,
+                            "printer-uri", NULL, printer_uri);
+              ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_NAME,
+                            "requesting-user-name", NULL, cupsUser ());
+              ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_NAME,
+              /* Translators: Name of job which makes printer to print test page */
+                            "job-name", NULL, _("Test page"));
+              response = cupsDoFileRequest (http, request, resource, filename);
+              httpClose (http);
+            }
+
+          g_free (filename);
+          g_free (printer_uri);
+          g_free (resource);
+        }
+      else
+        {
+          response = execute_maintenance_command (printer_name,
+                                                  "PrintSelfTestPage",
+          /* Translators: Name of job which makes printer to print test page */
+                                                  _("Test page"));
+        }
+
+      if (response)
+        {
+          if (response->state == IPP_ERROR)
+            g_warning ("An error has occured during printing of test page.");
+          ippDelete (response);
+        }
+    }
+}
+
 static void
 on_permission_changed (GPermission *permission,
                        GParamSpec  *pspec,
@@ -1976,11 +2140,15 @@ on_permission_changed (GPermission *permission,
 
   priv = PRINTERS_PANEL_PRIVATE (self);
 
-  is_authorized = g_permission_get_allowed (G_PERMISSION (priv->permission));
+  is_authorized = g_permission_get_allowed (G_PERMISSION (priv->permission)) &&
+    !g_settings_get_boolean (priv->lockdown_settings, "disable-print-setup");
 
   widget = (GtkWidget*) gtk_builder_get_object (priv->builder, "printer-add-button");
   gtk_widget_set_sensitive (widget, is_authorized);
 
+  widget = (GtkWidget*) gtk_builder_get_object (priv->builder, "printer-add-button2");
+  gtk_widget_set_sensitive (widget, is_authorized);
+
   widget = (GtkWidget*) gtk_builder_get_object (priv->builder, "printer-remove-button");
   gtk_widget_set_sensitive (widget, is_authorized);
 
@@ -1997,13 +2165,24 @@ on_permission_changed (GPermission *permission,
   gtk_widget_set_sensitive (widget, is_authorized);
 }
 
-enum
+static void
+on_lockdown_settings_changed (GSettings  *settings,
+                              const char *key,
+                              gpointer    user_data)
 {
-  NOTEBOOK_INFO_PAGE = 0,
-  NOTEBOOK_JOBS_PAGE,
-  NOTEBOOK_OPTIONS_PAGE,
-  NOTEBOOK_N_PAGES
-};
+  CcPrintersPanelPrivate  *priv;
+  CcPrintersPanel         *self = (CcPrintersPanel*) user_data;
+
+  if (g_str_equal (key, "disable-print-setup") == FALSE)
+    return;
+
+  priv = PRINTERS_PANEL_PRIVATE (self);
+
+  gtk_widget_set_sensitive (priv->lock_button,
+    !g_settings_get_boolean (priv->lockdown_settings, "disable-print-setup"));
+
+  on_permission_changed (priv->permission, NULL, user_data);
+}
 
 static void
 go_back_cb (GtkButton *button,
@@ -2113,6 +2292,10 @@ cc_printers_panel_init (CcPrintersPanel *self)
     gtk_builder_get_object (priv->builder, "printer-add-button");
   g_signal_connect (widget, "clicked", G_CALLBACK (printer_add_cb), self);
 
+  widget = (GtkWidget*)
+    gtk_builder_get_object (priv->builder, "printer-add-button2");
+  g_signal_connect (widget, "clicked", G_CALLBACK (printer_add_cb), self);
+
   widget = (GtkWidget*)
     gtk_builder_get_object (priv->builder, "printer-remove-button");
   g_signal_connect (widget, "clicked", G_CALLBACK (printer_remove_cb), self);
@@ -2139,7 +2322,7 @@ cc_printers_panel_init (CcPrintersPanel *self)
 
   widget = (GtkWidget*)
     gtk_builder_get_object (priv->builder, "print-test-page-button");
-  g_signal_connect (widget, "clicked", G_CALLBACK (printer_maintenance_cb), self);
+  g_signal_connect (widget, "clicked", G_CALLBACK (test_page_cb), self);
 
   widget = (GtkWidget*)
     gtk_builder_get_object (priv->builder, "back-button-1");
@@ -2157,6 +2340,12 @@ cc_printers_panel_init (CcPrintersPanel *self)
     gtk_builder_get_object (priv->builder, "printer-options-button");
   g_signal_connect (widget, "clicked", G_CALLBACK (switch_to_options_cb), self);
 
+  priv->lockdown_settings = g_settings_new ("org.gnome.desktop.lockdown");
+  g_signal_connect (priv->lockdown_settings,
+                    "changed",
+                    G_CALLBACK (on_lockdown_settings_changed),
+                    self);
+
 
   /* Set junctions */
   widget = (GtkWidget*)
@@ -2192,7 +2381,7 @@ cc_printers_panel_init (CcPrintersPanel *self)
 
   /* Add unlock button */
   priv->permission = (GPermission *)polkit_permission_new_sync (
-    "org.opensuse.cupspkhelper.mechanism.printeraddremove", NULL, NULL, NULL);
+    "org.opensuse.cupspkhelper.mechanism.all-edit", NULL, NULL, NULL);
   if (priv->permission != NULL)
     {
       widget = cc_lock_button_new (priv->permission);
